@@ -8,12 +8,12 @@ from pelecanus import PelicanJson
 
 from .entity_mapper import EntityMapper
 from .fact import Fact
-from .parsers import AddressParser, ContactEmailNamePairParser, ContactEmailParser, ContactPhoneParserHighPrecision, ContactPhoneParserHighRecall, ContactPhoneParserStrict
+from .parsers import AddressParser, ContactEmailNamePairParser, ContactEmailParser, ContactPhoneParserHighPrecision, \
+    ContactPhoneParserHighRecall, ContactPhoneParserStrict
 from .russian_transliterator import Transliterate
 
 
-# transliterator
-transliterate = Transliterate()
+russian_transliterator = Transliterate()
 
 
 class Document:
@@ -43,7 +43,8 @@ class Document:
             self,
             original_text: str,
             dominant_language_code: str,
-            stanza_words,
+            analysis_lang: str,
+            stanza_sentences: [list], 
             stanza_entities,
             concat_resources: dict,
             entity_mapper: Optional[EntityMapper] = None,
@@ -56,10 +57,12 @@ class Document:
         self.doc_path = doc_path
         self.analyzers = analyzers
         self.dominant_language_code = dominant_language_code
+        self.analysis_lang = analysis_lang
         self.json_doc = json_doc
 
         self.entity_mapper = entity_mapper
-        self.stanza_words = stanza_words
+        self.stanza_sentences = stanza_sentences
+        self.stanza_words = [word for sentence in self.stanza_sentences for word in sentence]
         self.stanza_entities = stanza_entities
 
         self.concat_resources = concat_resources
@@ -74,6 +77,15 @@ class Document:
         self.words()
 
 
+    def __get_doc_path(self, field: str) -> str:
+        """
+        :param field: Whether the doc_path uses the text or lemmas field.
+        :return: MLP representation of the doc_path
+        """
+        content = f"{self.doc_path}_mlp.{field}"
+        return content
+
+
     def __handle_existing_facts(self):
         """
         Add existing texta_facts inside the document into the private
@@ -86,7 +98,8 @@ class Document:
                 self.add_fact(fact)
 
 
-    def __remove_duplicate_facts(self, facts: List[dict]):
+    @staticmethod
+    def remove_duplicate_facts(facts: List[dict]):
         if facts:
             set_of_jsons = {json.dumps(fact, sort_keys=True, ensure_ascii=False) for fact in facts}
             without_duplicates = [json.loads(unique_fact) for unique_fact in set_of_jsons]
@@ -97,7 +110,7 @@ class Document:
 
     def facts_to_json(self) -> dict:
         facts = [fact.to_json() for fact in self.__texta_facts]
-        unique_facts = self.__remove_duplicate_facts(facts)
+        unique_facts = Document.remove_duplicate_facts(facts)
         return {"texta_facts": unique_facts}
 
 
@@ -105,11 +118,14 @@ class Document:
         self.__texta_facts.append(fact)
 
 
-    def document_to_json(self):
+    def document_to_json(self, use_default_doc_path=True) -> dict:
+        """
+        :param use_default_doc_path: Normal string values will be given the default path for facts but for dictionary input you already have them.
+        """
         list_of_path_keys = self.doc_path.split(".")
         root_key = "{}_mlp".format(list_of_path_keys[-1])
         path_to_mlp = list_of_path_keys[:-1] + [root_key] if len(list_of_path_keys) > 1 else [root_key]
-        mlp_result = self.to_json()
+        mlp_result = self.to_json(use_default_doc_path)
         nested_dict_wrapper = PelicanJson(self.json_doc)
         nested_dict_wrapper.set_nested_value(path_to_mlp, mlp_result["text"], force=True)
         nested_dict_wrapper.set_nested_value(["texta_facts"], mlp_result["texta_facts"], force=True)
@@ -117,38 +133,57 @@ class Document:
         return nested_dict_wrapper.convert()
 
 
-    def to_json(self) -> dict:
+    def to_json(self, use_default_doc_path=True) -> dict:
         container = dict()
-        container["text"] = self.get_words()
+        container["text"] = self.get_words(ssplit="sentences" in self.analyzers)
         texta_facts = self.facts_to_json()
-        container["lang"] = self.dominant_language_code
+        container["language"] = {"detected": self.dominant_language_code,
+                             "analysis": self.analysis_lang}
         if "lemmas" in self.analyzers: container["lemmas"] = self.get_lemma()
         if "pos_tags" in self.analyzers: container["pos_tags"] = self.get_pos_tags()
         # if "sentiment" in self.analyzers: container["sentiment"] = self.get_sentiment()
-        if "transliteration" in self.analyzers and self.__transliteration: container["transliteration"] = self.get_transliteration()
-        for fact in texta_facts["texta_facts"]:
-            fact["doc_path"] = "text.text"
+        if "transliteration" in self.analyzers and self.__transliteration: container[
+            "transliteration"] = self.get_transliteration()
+        if use_default_doc_path:
+            for fact in texta_facts["texta_facts"]:
+                fact["doc_path"] = "text.text"
         return {"text": container, **texta_facts}
 
 
     def lemmas(self):
-        self.__lemmas = [word.lemma.replace("_", "") if word and word.lemma else "X" for word in self.stanza_words]
+        for sent in self.stanza_sentences:
+            self.__lemmas.append([word.lemma.replace("_", "") if word and word.lemma else "X" for word in sent])
 
 
     def get_lemma(self) -> str:
-        return " ".join([a.strip() for a in self.__lemmas])
+        sentences = []
+        for sent_lemmas in self.__lemmas:
+            sentences.append(" ".join([a.strip() for a in sent_lemmas]))
+
+        if "sentences" in self.analyzers:
+            return "\n".join(sentences)
+        else:
+            return " ".join(sentences)
 
 
     def words(self):
-        self.__words = [word.text for word in self.stanza_words]
+        for sent in self.stanza_sentences:
+            self.__words.append([word.text for word in sent])
+
+    def sentences(self):
+        pass
 
 
-    def get_words(self) -> str:
-        return " ".join(self.__words)
+    def get_words(self, ssplit = False) -> str:
+        if ssplit:
+            return "\n".join([" ".join(sent_words) for sent_words in self.__words])
+        else:
+            return " ".join([" ".join(sent_words) for sent_words in self.__words])
 
 
     def pos_tags(self):
-        self.__pos_tags = [word.xpos if word and word.xpos and word.xpos != "_" else "X" if word.xpos == "_" else "X" for word in self.stanza_words]
+        self.__pos_tags = [word.xpos if word and word.xpos and word.xpos != "_" else "X" if word.xpos == "_" else "X"
+                           for word in self.stanza_words]
 
 
     def get_pos_tags(self) -> str:
@@ -164,21 +199,31 @@ class Document:
 
         hits = self.entity_mapper.map_entities(text)
         lemma_hits = self.entity_mapper.map_entities(lemmas, entity_types=["CURRENCY"])
-        # combine hits
-        all_hits = z = {**hits, **lemma_hits}
+
         # make facts
-        for entity_type, entity_values in all_hits.items():
+        for entity_type, entity_values in hits.items():
             for entity_value in entity_values:
                 new_fact = Fact(
                     fact_type=entity_type,
                     fact_value=entity_value["value"],
-                    doc_path=self.doc_path,
+                    doc_path=self.__get_doc_path("text"),
                     spans=[[entity_value["span"][0], entity_value["span"][1]]]
                 )
                 self.__texta_facts.append(new_fact)
+
+        for entity_type, entity_values in lemma_hits.items():
+            for entity_value in entity_values:
+                new_fact = Fact(
+                    fact_type=entity_type,
+                    fact_value=entity_value["value"],
+                    doc_path=self.__get_doc_path("lemmas"),
+                    spans=[[entity_value["span"][0], entity_value["span"][1]]]
+                )
+                self.__texta_facts.append(new_fact)
+
         # declare the entities processed
         self.entities_processed = True
-        
+
 
     def currency_sum(self):
         """
@@ -201,40 +246,43 @@ class Document:
                         new_fact = Fact(
                             fact_type="CURRENCY_SUM",
                             fact_value=fact_value,
-                            doc_path=self.doc_path,
+                            doc_path=self.__get_doc_path("text"),
                             spans=[match.start(), match.end()]
                         )
                         self.__texta_facts.append(new_fact)
-    
+
 
     def emails(self):
         text = self.get_words()
         emails = ContactEmailParser(text).parse()
-        self.__texta_facts.extend((email.to_fact(Document.FACT_NAME_EMAIL, self.doc_path) for email in emails))
+        self.__texta_facts.extend((email.to_fact(Document.FACT_NAME_EMAIL, self.__get_doc_path("text")) for email in emails))
 
 
     def phone_strict(self):
         text = self.get_words()
         phone_numbers_strict = ContactPhoneParserStrict(text).parse()
-        self.__texta_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_STRICT, self.doc_path) for number in phone_numbers_strict))
+        self.__texta_facts.extend(
+            (number.to_fact(Document.FACT_NAME_PHONE_STRICT, self.__get_doc_path("text")) for number in phone_numbers_strict))
 
 
     def phone_high_recall(self):
         text = self.get_words()
         phone_numbers = ContactPhoneParserHighRecall(text, months=self.concat_resources["months"]).parse()
-        self.__texta_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.doc_path) for number in phone_numbers))
+        self.__texta_facts.extend(
+            (number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.__get_doc_path("text")) for number in phone_numbers))
 
 
     def phone_high_precision(self):
         text = self.get_words()
         phone_numbers_high_precision = ContactPhoneParserHighPrecision(text).parse()
-        self.__texta_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_PRECISION, self.doc_path) for number in phone_numbers_high_precision))
+        self.__texta_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_PRECISION, self.__get_doc_path("text")) for number in
+                                   phone_numbers_high_precision))
 
 
     def addresses(self):
         text = self.get_words()
         addresses = AddressParser(text, self.stanza_entities, self.dominant_language_code).parse()
-        self.__texta_facts.extend((addr.to_fact(Document.FACT_NAME_ADDRESS, self.doc_path) for addr in addresses))
+        self.__texta_facts.extend((addr.to_fact(Document.FACT_NAME_ADDRESS, self.__get_doc_path("text")) for addr in addresses))
 
 
     def transliteration(self):
@@ -249,7 +297,7 @@ class Document:
 
     @staticmethod
     def _transliterate_russian_word(word):
-        translit_word = transliterate([word.text.strip()])
+        translit_word = russian_transliterator([word.text.strip()])
         try:
             translit_word = translit_word[0].strip()
         except IndexError:
@@ -320,7 +368,7 @@ class Document:
                     new_fact = Fact(
                         fact_type=entity.type,
                         fact_value=entity.text,
-                        doc_path=self.doc_path,
+                        doc_path=self.__get_doc_path("text"),
                         spans=[best_matching_span]
                     )
                     self.__texta_facts.append(new_fact)
@@ -333,7 +381,8 @@ class Document:
         """
         text = self.get_words()
         email_name_pairs = ContactEmailNamePairParser(text).parse()  # bounded -> str "name mail"
-        self.__texta_facts.extend((emailpair.to_fact(Document.FACT_NAME_NAMEMAIL, self.doc_path) for emailpair in email_name_pairs))
+        self.__texta_facts.extend(
+            (emailpair.to_fact(Document.FACT_NAME_NAMEMAIL, self.__get_doc_path("text")) for emailpair in email_name_pairs))
 
 
     def remove_duplicate_facts_by_span(self, facts):
@@ -357,7 +406,7 @@ class Document:
                 name_fact = Fact(
                     fact_type="PER",
                     fact_value=name_fact_value.strip("><\)\(:;-\.,\!\?"),
-                    doc_path=self.doc_path,
+                    doc_path=self.__get_doc_path("text"),
                     spans=[(fact.spans[0][0], fact.spans[0][0] + len(name_fact_value))]
                 )
 
@@ -365,7 +414,7 @@ class Document:
                 email_fact = Fact(
                     fact_type="EMAIL",
                     fact_value=email_fact_value.strip("><\)\(:;-\.,\!\?"),
-                    doc_path=self.doc_path,
+                    doc_path=self.__get_doc_path("text"),
                     spans=[(fact.spans[0][1] - len(email_fact_value), fact.spans[0][1])]
                 )
 
@@ -374,7 +423,8 @@ class Document:
                 ends += [email_fact.spans[0][1], name_fact.spans[0][1]]
 
         for fact in facts:
-            if fact.fact_type != Document.FACT_NAME_NAMEMAIL and fact.spans[0][0] not in starts and fact.spans[0][1] not in ends:
+            if fact.fact_type != Document.FACT_NAME_NAMEMAIL and fact.spans[0][0] not in starts and fact.spans[0][
+                1] not in ends:
                 fact.fact_value = fact.fact_value.strip("><\)\(:;-\.,\!\?")
                 new_facts += [fact]
         new_facts_values = list()
@@ -399,7 +449,9 @@ class Document:
             issubitem = False
             for j, item2 in enumerate(bound.items()):
                 if i != j and i not in no_is and j not in no_is:
-                    if re.search("(^|\s)" + re.escape(item1[0].strip("><)(:;-.,!?")) + "(\s|$|\@[a-z]{4,5}\.[a-z]{2})", item2[0].strip("><)(:;-.,!?")) and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(item2[0].strip("><\)\(:;-\.,\!\?")):
+                    if re.search("(^|\s)" + re.escape(item1[0].strip("><)(:;-.,!?")) + "(\s|$|\@[a-z]{4,5}\.[a-z]{2})",
+                                 item2[0].strip("><)(:;-.,!?")) and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(
+                        item2[0].strip("><\)\(:;-\.,\!\?")):
                         if item2[1] in new_bound:
                             new_bound[item2[1]] += [item2[0].strip("><)(:;-.,!?")]
                         else:
@@ -409,10 +461,12 @@ class Document:
                     else:  # each word of this entity exists also in the other one.
                         notsameparts = False
                         for part_of_entity in item1[0].split(" "):
-                            if not re.search("(^|\s)" + re.escape(part_of_entity.strip("><)(:;-.,!?")) + "(\s|$)", item2[0].strip("><)(:;-.,!?")):
+                            if not re.search("(^|\s)" + re.escape(part_of_entity.strip("><)(:;-.,!?")) + "(\s|$)",
+                                             item2[0].strip("><)(:;-.,!?")):
                                 notsameparts = True
                                 break
-                        if not notsameparts and len(item1[0].strip("><)(:;-.,!?")) <= len(item2[0].strip("><)(:;-.,!?")):
+                        if not notsameparts and len(item1[0].strip("><)(:;-.,!?")) <= len(
+                                item2[0].strip("><)(:;-.,!?")):
                             if item2[1] in new_bound:
                                 new_bound[item2[1]] += [re.sub("(\'|\")", "*", item2[0].strip("><)(:;-.,!?"))]
                             else:
@@ -442,7 +496,8 @@ class Document:
         if re.search(self.concat_resources["space_between_not_ok"], text_between):
             return False
         for month in self.concat_resources["months"]:
-            if re.search("[0-9]{1,2} ?" + month + " ?\.? ?[0-9]{4}( г\.)? ?, ?в? [0-9]{2} ?: ?[0-9]{2}", text_between.lower()):
+            if re.search("[0-9]{1,2} ?" + month + " ?\.? ?[0-9]{4}( г\.)? ?, ?в? [0-9]{2} ?: ?[0-9]{2}",
+                         text_between.lower()):
                 return False
         return True
 
@@ -462,7 +517,8 @@ class Document:
         Bound facts found in that email, that has distance of  <=self.CLOSE_FACT_DISTANCE and that has self.space_between_ok True, together.
         """
         facts = sorted(facts, key=lambda item: item.spans[0][0])
-        new_bound = {'doc_path': self.doc_path, 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {}, "str_values": []}
+        new_bound = {'doc_path': self.__get_doc_path("text"), 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {},
+                     "str_values": []}
         new_bounded = []
         for fact in facts:
             if len(new_bound["spans"]) == 0:
@@ -470,7 +526,9 @@ class Document:
                 new_bound["str_val"][fact.fact_type] = [fact.fact_value]
                 new_bound["str_values"] = [(fact.fact_value, fact.fact_type)]
             else:
-                if abs(fact.spans[0][0] - new_bound["spans"][-1][-1]) < self.CLOSE_FACT_DISTANCE and self.space_between_ok(self.get_words(), new_bound["spans"][-1], fact.spans[0]):
+                if abs(fact.spans[0][0] - new_bound["spans"][-1][
+                    -1]) < self.CLOSE_FACT_DISTANCE and self.space_between_ok(self.get_words(), new_bound["spans"][-1],
+                                                                              fact.spans[0]):
                     new_bound["spans"] += fact.spans
                     new_bound["str_values"] += [(fact.fact_value, fact.fact_type)]
                     if fact.fact_type in new_bound["str_val"]:
@@ -480,7 +538,8 @@ class Document:
                 else:  # start a new bound
                     if len(new_bound["spans"]) > 1:
                         new_bounded += [new_bound]
-                    new_bound = {'doc_path': self.doc_path, 'fact': 'BOUNDED', 'spans': [], "str_val": {}, "str_values": []}
+                    new_bound = {'doc_path': self.__get_doc_path("text"), 'fact': 'BOUNDED', 'spans': [], "str_val": {},
+                                 "str_values": []}
                     new_bound["spans"] += fact.spans
                     new_bound["str_values"] = [(fact.fact_value, fact.fact_type)]
                     new_bound["str_val"][fact.fact_type] = [fact.fact_value]
@@ -507,22 +566,27 @@ class Document:
                             to_delete_is += [j]
                         elif span1[1] < span2[1]:
                             to_delete_is += [i]
-            new_bound_erased_coinciding = {'doc_path': self.doc_path, 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {}}
+            new_bound_erased_coinciding = {'doc_path': self.__get_doc_path("text"), 'fact': Document.FACT_NAME_BOUNDED, 'spans': [],
+                                           'str_val': {}}
             for i, span1 in enumerate(new_bound["spans"]):
                 if i not in to_delete_is:
                     new_bound_erased_coinciding["spans"] += [span1]
 
                     if new_bound["str_values"][i][1] in new_bound_erased_coinciding["str_val"]:
-                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] += [new_bound["str_values"][i][0]]
+                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] += [
+                            new_bound["str_values"][i][0]]
                     else:
-                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] = [new_bound["str_values"][i][0]]
-            new_bound_erased_coinciding["str_val"] = self.clean_similar_in_strval(new_bound_erased_coinciding["str_val"])
+                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] = [
+                            new_bound["str_values"][i][0]]
+            new_bound_erased_coinciding["str_val"] = self.clean_similar_in_strval(
+                new_bound_erased_coinciding["str_val"])
             hasmanyentities = False
             for entity_type in new_bound_erased_coinciding["str_val"]:
                 if len(new_bound_erased_coinciding["str_val"][entity_type]) > 1:
                     hasmanyentities = True
                     break
-            if len(new_bound_erased_coinciding["spans"]) > 1 and (hasmanyentities or len(new_bound_erased_coinciding["str_val"].items()) > 1):
+            if len(new_bound_erased_coinciding["spans"]) > 1 and (
+                    hasmanyentities or len(new_bound_erased_coinciding["str_val"].items()) > 1):
                 new_bounded_coinciding_erased += [new_bound_erased_coinciding]
         return new_bounded_coinciding_erased
 
@@ -537,7 +601,8 @@ class Document:
         for i, new_bound in enumerate(new_bounded):
             issubset = False
             for j, another_bound in enumerate(new_bounded):
-                if j not in concatenated_bound_is and i not in concatenated_bound_is and i != j and set(self.key_value_single_pairs(new_bound["str_val"])) <= set(self.key_value_single_pairs(another_bound["str_val"])):
+                set_comparison = set(self.key_value_single_pairs(new_bound["str_val"])) <= set(self.key_value_single_pairs(another_bound["str_val"]))
+                if j not in concatenated_bound_is and i not in concatenated_bound_is and i != j and set_comparison:
                     issubset = True
                     another_bound["spans"] += new_bound["spans"]
                     concatenated_bound_is += [i]
@@ -556,7 +621,7 @@ class Document:
                     fact_type=new_bound["fact"],
                     fact_value=str(new_bound["str_val"]),  ##add lemmas
                     fact_lemma=str(lemmas),
-                    doc_path=self.doc_path,
+                    doc_path=self.__get_doc_path("text"),
                     spans=new_bound["spans"]
                 )
                 new_bounds.append(new_fact)
@@ -569,16 +634,18 @@ class Document:
         remove previous phone facts
         add phoneparser made for emails (ContactPhoneParserHighRecall)
         """
-        new_facts = [fact for fact in facts if not re.match("PHONE|"+Document.FACT_NAME_BOUNDED, fact.fact_type)]
+        new_facts = [fact for fact in facts if not re.match("PHONE|" + Document.FACT_NAME_BOUNDED, fact.fact_type)]
         text = self.get_words()
         phone_numbers = ContactPhoneParserHighRecall(text, months=self.concat_resources["months"]).parse()
-        new_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.doc_path) for number in phone_numbers))
+        new_facts.extend(
+            (number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.__get_doc_path("text")) for number in phone_numbers))
         return new_facts
 
 
     def bounded(self):
         facts = self.remove_duplicate_facts_by_span(self.__texta_facts)
-        facts = self.remove_old_bounded_and_add_phone_high_recall(facts)  # bounded uses only ContactPhoneParserHighRecall results and doesn't look at old BOUNDED facts.
+        facts = self.remove_old_bounded_and_add_phone_high_recall(
+            facts)  # bounded uses only ContactPhoneParserHighRecall results and doesn't look at old BOUNDED facts.
         new_bounded = self.bound_close_ones(facts)
         new_bounded = self.remove_overlaping_in_bounded(new_bounded)
         new_bounded = self.concatenate_subset_bounds(new_bounded)
