@@ -1,10 +1,10 @@
 import logging
 import os
 import pathlib
+import shutil
 from typing import List
 from urllib.parse import urlparse
 from urllib.request import urlopen
-
 import regex as re
 import stanza
 from bs4 import BeautifulSoup
@@ -14,26 +14,31 @@ from pelecanus import PelicanJson
 from texta_mlp.document import Document
 from texta_mlp.entity_mapper import EntityMapper
 from texta_mlp.exceptions import LanguageNotSupported
+from texta_mlp.utils import parse_bool_env
+
 
 # Languages supported by default.
-DEFAULT_LANG_CODES = ("et", "ru", "en")
+DEFAULT_LANG_CODES = ("et", "ru", "en", "ar")
 
 # URLs for default Entity Mapper data sources.
 ENTITY_MAPPER_DATA_URLS = (
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/entity_mapper/addresses.json",
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/entity_mapper/companies.json",
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/entity_mapper/currencies.json"
+    "https://packages.texta.ee/texta-resources/entity_mapper/addresses.json",
+    "https://packages.texta.ee/texta-resources/entity_mapper/companies.json",
+    "https://packages.texta.ee/texta-resources/entity_mapper/currencies.json"
 )
 
 # URLs for Concatenator data sources.
 CONCATENATOR_DATA_FILES = (
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/concatenator/months.txt",
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/concatenator/not_entities.txt",
-    "https://git.texta.ee/texta/texta-resources/-/raw/master/concatenator/space_between_not_ok.txt",
+    "https://packages.texta.ee/texta-resources/concatenator/months.txt",
+    "https://packages.texta.ee/texta-resources/concatenator/not_entities.txt",
+    "https://packages.texta.ee/texta-resources/concatenator/space_between_not_ok.txt",
 )
 
 # Location of the resource dir where models are downloaded
-DEFAULT_RESOURCE_DIR = os.path.join(os.getcwd(), "data")
+DEFAULT_RESOURCE_DIR = os.getenv("TEXTA_MLP_DATA_DIR", os.path.join(os.getcwd(), "data"))
+
+# Data refresh means deleting all existing models and downloading new ones
+REFRESH_DATA = parse_bool_env("TEXTA_MLP_REFRESH_DATA", False)
 
 # List of all analyzers supported by MLP
 SUPPORTED_ANALYZERS = (
@@ -47,7 +52,8 @@ SUPPORTED_ANALYZERS = (
     "entities",
     "namemail",
     "bounded",
-    "currency_sum"
+    "currency_sum",
+    "sentences"
 )
 
 # Here we define languages with NER support to avoid Stanza trying to load them for languages without NER support.
@@ -65,24 +71,22 @@ class MLP:
             use_default_language_code=True,
             resource_dir: str = DEFAULT_RESOURCE_DIR,
             logging_level="error",
-            use_gpu=True
+            use_gpu=True,
+            refresh_data=REFRESH_DATA
     ):
         self.supported_langs = language_codes
         self.logger = logging.getLogger()
         self.default_lang = default_language_code
         self.use_default_lang = use_default_language_code
-        self.langs_to_transliterate = ["ru"]
         self.resource_dir = resource_dir
-        self.use_gpu=use_gpu
+        self.use_gpu = use_gpu
 
         self.resource_dir_pathlib = pathlib.Path(resource_dir)
         self.not_entities_path = self.resource_dir_pathlib / "concatenator" / "not_entities.txt"
         self.space_between_not_ok_path = self.resource_dir_pathlib / "concatenator" / "space_between_not_ok.txt"
         self.months_path = self.resource_dir_pathlib / "concatenator" / "months.txt"
 
-        self.download_stanza_resources(self.resource_dir, self.supported_langs, logger=self.logger)
-        self.download_entity_mapper_resources(self.resource_dir, logger=self.logger)
-        self.download_concatenator_resources(self.resource_dir, logger=self.logger)
+        self.prepare_resources(refresh_data)
 
         self.stanza_pipelines = self._load_stanza_pipelines(logging_level)
         self.entity_mapper = self._load_entity_mapper()
@@ -98,6 +102,20 @@ class MLP:
         }
 
 
+    def prepare_resources(self, refresh_data):
+        """
+        Prepares all resources for MLP.
+        """
+        # delete data if refresh asked
+        if refresh_data:
+            shutil.rmtree(self.resource_dir)
+            self.logger.info("MLP data directory deleted.")
+        # download resources
+        self.download_stanza_resources(self.resource_dir, self.supported_langs, logger=self.logger)
+        self.download_entity_mapper_resources(self.resource_dir, logger=self.logger)
+        self.download_concatenator_resources(self.resource_dir, logger=self.logger)
+
+
     @staticmethod
     def download_stanza_resources(resource_dir: str, supported_langs: List[str], logger=None):
         """
@@ -165,12 +183,6 @@ class MLP:
             lang = detect(text)
         except:
             lang = None
-        # check if detected language is supported
-        if lang not in self.supported_langs:
-            if self.use_default_lang:
-                lang = self.default_lang
-            else:
-                raise LanguageNotSupported("Detected language is not supported: {}.".format(lang))
         return lang
 
 
@@ -179,13 +191,22 @@ class MLP:
         # detect language
         if not lang:
             lang = self.detect_language(processed_text)
-
-        words, entities = self._get_stanza_tokens(lang, processed_text) if processed_text else ([], [])
+        '''
+        check if detected language is supported if the language is not supported it will use default_lang to load
+        stanza models yet keep the document lang as the detected language
+        '''
+        if lang not in self.supported_langs:
+            analysis_lang = self.default_lang
+            sentences, entities = self._get_stanza_tokens(analysis_lang, processed_text) if processed_text else ([], [])
+        else:
+            analysis_lang = lang
+            sentences, entities = self._get_stanza_tokens(analysis_lang, processed_text) if processed_text else ([], [])
 
         document = Document(
             original_text=processed_text,
             dominant_language_code=lang,
-            stanza_words=words,
+            analysis_lang=analysis_lang,
+            stanza_sentences=sentences,
             stanza_entities=entities,
             analyzers=analyzers,
             json_doc=json_object,
@@ -210,23 +231,23 @@ class MLP:
         if lang == "ru":
             raw_text = pat.sub("_", raw_text)
 
-        # Separate all the words into a separate list.
-
         pipeline = self.stanza_pipelines[lang](raw_text)
 
-        words = []
+        sentences = []
         entities = []
         pip_pat = re.compile(r"(?<=\d)_(?=\d)")
         for sentence in pipeline.sentences:
+            words = []
             for word in sentence.words:
                 # Russian HACK (2/2)
                 # replaces back "#" to "-" between digits.
                 if lang == "ru":
                     word.text = pip_pat.sub("-", word.text)
                 words.append(word)
+            sentences.append(words)
             for entity in sentence.entities:
                 entities.append(entity)
-        return words, entities
+        return sentences, entities
 
 
     def _get_stanza_ner(self, lang: str, raw_text: str):
@@ -249,7 +270,7 @@ class MLP:
                 logging_level=logging_level
             )
         return stanza_pipelines
-    
+
 
     @staticmethod
     def _get_stanza_processors(lang):
@@ -268,7 +289,7 @@ class MLP:
         Processes raw text.
         :param: raw_text str: Text to be processed.
         :param: analyzers list: List of analyzers to be used.
-        :return: Processed text as document ready for Elastic.       
+        :return: Processed text as document ready for Elastic.
         """
         loaded_analyzers = self._load_analyzers(analyzers, SUPPORTED_ANALYZERS)
         document = self.generate_document(raw_text, loaded_analyzers, lang=lang)
@@ -330,10 +351,22 @@ class MLP:
                             # For every analyzer, activate the function that processes it from the
                             # document class.
                             getattr(doc, analyzer)()
-                        data = doc.document_to_json()
-                        container.append(data)
-                    else:
-                        container.append(None)
+
+                        result = doc.document_to_json(use_default_doc_path=False)
+                        new_facts = result.pop("texta_facts", [])
+                        existing_facts = document.get("texta_facts", [])
+                        unique_facts = Document.remove_duplicate_facts(new_facts + existing_facts)
+                        result["texta_facts"] = unique_facts
+                        document = result
+
+            if document:
+                # Add in texta_facts even if nothing was done due to missing values.
+                facts = document.get("texta_facts", [])
+                document["texta_facts"] = facts
+                container.append(document)
+            else:
+                # Add in at least something to avoid problems with operations that include indexing.
+                container.append({})
 
         return container
 
